class CLEAN_TAGS
    --ad hoc reductions
inherit
    OPTIMIZER;
    TAG_SINGLETON
creation make
feature
    name : STRING is "tag cleaner"

    last_row : ROW
	--holds last nonempty row or Void

    change : BOOLEAN

    execute is
	--eliminate sequences like <B><1></1></B>
	--also handle multiple <EM>'s
    local
	i : INTEGER
	row : ROW
	done : BOOLEAN
    do
	from
	until done
	loop

	    change := false
	    --a little messy because we get rid of rows with
	    --empty tag_lists at the same time
	    last_row := Void
	    from  i := program.lower
	    until i > program.upper
	    loop
		row := program @ i
		if clean_tag_list(row) then
		    program.remove(i)
		else
		    i := i + 1
		end
	    end
	    if not change then
		done := true
	    end
	end
    end
    
    clean_tag_list(row : ROW) : BOOLEAN is
	--return true if tag_list ends up empty
	--get rid of <B></B> et. al
	--get rid of <EM><EM>
	--do <6><7> ... </7></6> --> <7>...</7>
	--do <r><g> ... </g></r> --> <g>...</g>
	--do <B><PL> ... </PL></B> --> <PL>...</PL>
	--do <EM><S> ... </S></EM> --> <S>...</S>
    local
	it : ITERATOR[TAG_HOLDER]
	t, p : TAG_HOLDER
	tag : TAG
	newl : LINKED_LIST[TAG_HOLDER]
	last_em : TAG_HOLDER
	last_color : TAG_HOLDER
	last_size : TAG_HOLDER

	other_tags : LINKED_LIST[TAG_HOLDER]
	ems_tags : LINKED_LIST[TAG_HOLDER]
	list : LINKED_LIST[TAG_HOLDER]

	old_count : INTEGER
    do
	list := row.tag_list
	old_count := list.count

	remove_palindrome(list)

	!!other_tags.make
	!!ems_tags.make

	it := list.get_new_iterator
	from  it.start
	until it.is_off
	loop
	    t := it.item
	    p := t.partner
	    tag := t.tag
	    if tag.tag_start then

		--eliminate COLOR_TAG and SIZE_TAG shadows
		if is_start_color_tag(tag) then
		    if last_color /= Void then
			check_shadow(last_color, t)
		    end
		    last_color := t
		elseif is_start_size_tag(tag) then
		    if last_size /= Void then
			check_shadow(last_size, t)
		    end
		    last_size := t
		else
		    other_tags.add_last(t)
		    if tag = em_tag then
    --can eliminate same row <EM>'s as long as there aren't <PL>'s
    --see PL section
    --(they have to end on the same row too!)
			if last_em /= Void then
			    if last_em.partner.row = p.row then
				last_em.set_redundant
				remove_tag_holder(last_em.partner)
				t.set_redundant
				remove_tag_holder(p)
				last_em := Void
			    else
				last_em := t
			    end
			else
			    last_em := t
			end
			ems_tags.add_last(t)
		    elseif tag = s_tag then
			--S can shadow EM
			ems_tags.add_last(t)
			last_em := Void
			trim(ems_tags)
			ems_tags.clear
		    elseif tag = pl_tag then
			--PL can shadow all non-size/color tags
			last_em := Void
			trim(other_tags)
		    end
		end
	    end
	    it.next
	end

	--by now, in a row,
	--there are no end tags after a start tag (since tags must be nested)
	--for each category except U there is at most one tag
	--(categories are colour, size, B, I, U, EM, S, TT)
	--an EM tag never follows a S tag

	!!newl.make
	it := list.get_new_iterator
	from it.start
	until it.is_off
	loop
	    t := it.item
	    if not t.redundant then
		newl.add_last(t)
	    end
	    it.next
	end

	list := newl

	row.put_tag_list(list)
	if list.count < old_count then
	    change := true
	end

	if list.is_empty and then last_row /= Void then
	    last_row.text.append(row.text)
	    Result := true
	else
	    last_row := row
	end
    end

    check_shadow(t1, t2 : TAG_HOLDER) is
	-- if t1 is shadowed by t2 mark it redundant
    local
	t1p : TAG_HOLDER
    do
	t1p := t1.partner
	if t1p.row = t2.partner.row then
	    t1.set_redundant
	    remove_tag_holder(t1p)
	end
    end

    trim(list : LINKED_LIST[TAG_HOLDER]) is
	--if any other elements are shadowed by the last element
	--then mark them redundant
    require
	not list.is_empty
    local
	done : BOOLEAN
	lastprow : ROW
	t, p : TAG_HOLDER
    do
	lastprow := list.last.partner.row
	from
	until done
	loop
	    t := list.first
	    if t = list.last then
		done := true
	    else
		list.remove_first
		p := t.partner
		if p.row = lastprow then
		    t.set_redundant
		    remove_tag_holder(p)
		end
	    end
	end
    end

    remove_palindrome(tag_list : LINKED_LIST[TAG_HOLDER]) is
	--get rid of </I></B><B><I> etc.
    local
	i, j : INTEGER
	t, t0 : TAG_HOLDER
	done : BOOLEAN
    do
	--find first starting tag
	from i := tag_list.lower
	until done or else i > tag_list.upper
	loop
	    t := tag_list @ i
	    if t.tag.tag_start then
		done := True
	    else
		i := i + 1
	    end
	end

	--start removing </tag><tag> pairs from the inside out
	if done then
	    done := false
	    j := i - 1
	    from
	    until done
	    loop
		if j < tag_list.lower or else i > tag_list.upper then
		    done := true
		else
		    t := tag_list @ i
		    t0 := tag_list @ j
		    if t0.tag = opposite_tag(t.tag) then
			t0.partner.partner_with(t.partner)
			tag_list.remove(i)
			tag_list.remove(j)
			i := i - 1
			j := j - 1
		    else
			done := true
		    end
		end
	    end
	end
    end

    remove_tag_holder(th : TAG_HOLDER) is
    local
	tl : LINKED_LIST[TAG_HOLDER]
	i : INTEGER
    do
	tl := th.row.tag_list

	i := tl.fast_index_of(th)
	if i <= tl.upper then
	    tl.remove(i)
	end
    end
end
