deferred class PL_SHORTCUT
inherit OPTIMIZER
feature
    pl_shortcut_optimize is
	--try to call this after a merge
    local
	it : ITERATOR[NODE]
	cl : CLUSTER
    do
	pl_shortcut_children_and_absorb
	it := children.get_new_iterator
	from it.start
	until it.is_off
	loop
	    cl ?= it.item
	    if cl /= Void then
		cl.pl_shortcut_optimize
	    end
	    it.next
	end
    end

    pl_shortcut_children_and_absorb is
    local
	i : INTEGER
	cl : CLUSTER
	ct : CLUSTER_TEXT
	ignore : BOOLEAN
    do
	from i := children.lower + 1
	until i >= children.upper
	loop
	    cl ?= children @ i
	    if cl /= Void then
		if try_pl_shortcut_case2(i) then
		    --compensate for losing current child
		    i := i - 1
		end
	    else
		ct ?= children @ i
		if try_pl_shortcut(i) then
		    --compensate for losing current child
		    i := i - 1
		end
	    end
	    i := i + 1
	end
	ignore := try_absorb
    end

    try_pl_shortcut(i : INTEGER) : BOOLEAN is
    local
	c1, c2 : CLUSTER
	ct : CLUSTER_TEXT
    do
	c1 ?= children @ (i - 1)
	ct ?= children @ i
	c2 ?= children @ (i + 1)
	if c1 /= Void and then c2 /= Void then
	    pl_common_cluster(ct, c1, c2)
	    if pl_outer_cover.cost > pl_inner_cover.cost then
		--worth doing
		do_pl_shortcut(ct, c1, c2, i - 1)
	    end
	end
    end

    try_pl_shortcut_case2(i : INTEGER) : BOOLEAN is
    local
	c1, c2 : CLUSTER
	cl : CLUSTER
	cover : CLUSTER
	ct : CLUSTER_TEXT
	done : BOOLEAN
    do
	c1 ?= children @ (i - 1)
	cl ?= children @ i
	c2 ?= children @ (i + 1)
	if c1 /= Void and then c2 /= Void then
	    if cl.children.count = 1 then
		ct ?= cl.children.first
		if ct /= Void then
		    pl_common_cluster(ct, c1, c2)
		    if pl_outer_cover.cost + cl.cost > pl_inner_cover.cost then
			--worth doing
			do_pl_shortcut(ct, c1, c2, i - 1)
			done := true
		    end
		end
	    end
	    if not done and then cl.pl_flag then
		-- can do another trick
		--cl should already match the context
		cover := c1.common_cluster(c2)
		cover.put_size_color(Void, Void)
		--too complicated to handle color/size here
		--will be done by nesting anyway
		if cover.cost > 0 then
		    --worth doing
		    c1.subtract(cover)
		    c2.subtract(cover)
		    cover.add_last_child(c1)
		    cover.add_last_child(cl)
		    cover.add_last_child(c2)
		    children.remove(i - 1)
		    children.remove(i - 1)
		    put_child(cover, i - 1)
		    set_change_flag
		end
	    end
	end
    end

    do_pl_shortcut(ct : CLUSTER_TEXT; c1, c2 : CLUSTER;
	    pos : INTEGER) is
    do
	c1.subtract(pl_outer_cover)
	c2.subtract(pl_outer_cover)
	pl_outer_cover.add_last_child(c1)
	pl_outer_cover.add_last_child(pl_inner_cover)
	pl_outer_cover.add_last_child(c2)
	pl_inner_cover.add_last_child(ct)
	children.remove(pos)
	children.remove(pos)
	put_child(pl_outer_cover, pos)
	set_change_flag
    end

    pl_inner_cover : CLUSTER
    pl_outer_cover : CLUSTER

    pl_common_cluster(ct : CLUSTER_TEXT; c1, c2 : like Current) is
	--special for the PL shortcut optimization
	--different pl's on either side break it
	--colors can't be saved unless c is colored
	--similarly for sizes
    local
	pc : CLUSTER
	c : CONTEXT
	common_color, common_size : TAG
    do
	!!pl_inner_cover.make
	!!pl_outer_cover.make
	c := ct.context
	pl_inner_cover.put_bisemttplusc(
		c.b_flag,
		c.i_flag,
		c.s_flag,
		c.em_flag,
		c.tt_flag,
		true,
		c.u_count,
		c.color,
		c.size)
--if pl_inner_cover.color /= c.color then
--BUG in SmallEiffel? there aren't equal!
--end

	--if the c isn't root colored then we're allowed to
	--lift common colors. Same goes for sizes
	if pl_inner_cover.color /= Void then
	    if c1.color = c2.color then
		common_color := c1.color
		if common_color = pl_inner_cover.color then
		    pl_inner_cover.put_color(Void)
		end
	    end
	end
	if pl_inner_cover.size /= Void then
	    if c1.size = c2.size then
		common_size := c1.size
		if common_size = pl_inner_cover.size then
		    pl_inner_cover.put_size(Void)
		end
	    end
	end
	if c1.pl_flag = c2.pl_flag then
	    pl_outer_cover.put_bisemttplusc(
		c1.b_flag and c2.b_flag,
		c1.i_flag and c2.i_flag,
		c1.s_flag and c2.s_flag,
		c1.em_flag and c2.em_flag,
		c1.tt_flag and c2.tt_flag,
		c1.pl_flag, --pl_flag and other.pl_flag,
		c1.u_count.min(c2.u_count),
		common_size,
		common_color)
	else
	    pl_outer_cover.put_size_color(common_size, common_color)
	end
	--one more thing; if there is no common color,
	--then the inner color might be right anyway
	--(same for sizes)
	if common_color = Void then
	    if pl_inner_cover.color /= Void then
		from pc := c1.parent
		until pc = Void or else pc.color /= Void
		loop
		    pc := pc.parent
		end
		if pc /= Void and then pc.color = pl_inner_cover.color then
		    pl_inner_cover.put_color(Void)
		end
	    end
	end
	if common_size = Void then
	    if pl_inner_cover.size /= Void then
		from pc := c1.parent
		until pc = Void or else pc.size /= Void
		loop
		    pc := pc.parent
		end
		if pc /= Void and then pc.size = pl_inner_cover.size then
		    pl_inner_cover.put_size(Void)
		end
	    end
	end
    end
end
