deferred class SIMPLIFY_WHITESPACE
inherit
    SML_HELPER;
    IS_WHITESPACE
feature
    simplify_whitespace is
    local
	i : INTEGER
	r1, r2 : ROW
	flag : BOOLEAN
    do
	--try to merge rows that are all whitespace with other rows
	i := program.lower
	from i := next_all_space(i)
	until i > program.upper
	loop
	    r1 := program @ i
	    flag := try_merge_all_space(i)
	    if not flag then
		i := i + 1
	    end
	    i := next_all_space(i)
	end

	--get rid of unnecessary spaces between rows
	from i := program.lower
	until i >= program.upper
	    --no need to try the last line because it's impossible
	    --that the last space is redundant
	    --also, it's the only one that could be empty so it
	    --saves us testing for emptiness in r1
	loop
	    r1 := program @ i
	    if is_whitespace(r1.text.last) then
		r2 := program @ (i+1)
		if space_match(r1.context, r2.context) then
		    if not r1.context.tt_flag and then not r2.text.is_empty and then is_whitespace(r2.text.first) then
			r2.text.remove_first(1)
		    end
		end
	    end

	    i := i + 1
	end
    end

    is_all_whitespace(s : STRING) : BOOLEAN is
    local
	i : INTEGER
    do
	Result := true
	from i := s.lower
	until not Result or else i > s.upper
	loop
	    if not is_whitespace(s.item(i)) then
		Result := false
	    else
		i := i + 1
	    end
	end
    end

    next_all_space(i : INTEGER) : INTEGER is
	--find next row i that is all whitespace and nonempty
    local
	r : ROW
	done : BOOLEAN
    do
	from Result := i
	until done or Result > program.upper
	loop
	    r := program @ Result
	    if r.context.tt_flag then
		if not r.text.is_empty and then is_all_whitespace(r.text) then
		    done := True
		else
		    Result := Result + 1
		end
	    else
		if r.text.count = 1 and then is_whitespace(r.text.first) then
		    done := True
		else
		    Result := Result + 1
		end
	    end
	end
    end
    
    try_merge_all_space(i : INTEGER) : BOOLEAN is
	--try to merge row i (which is all whitespace)
	--with previous or next row
    local
	r0, r1, r2 : ROW
    do
	r1 := program @ i
	if i > program.lower then
	    r0 := program @ (i-1)
	    if space_match(r0.context, r1.context) then
		if i = program.upper then
		    !!r2.make
		    r2.put_context(r1.context)
		    r2.put_text("")
		    program.add_last(r2)
		else
		    r2 := program @ (i+1)
		end
		r0.append_space_row(r1, r2)
		program.remove(i)
		Result := true
	    end
	end
	if not Result and i < program.upper then
		--try prepending to next row instead
	    r2 := program @ (i+1)
	    if space_match(r1.context, r2.context) then
		r2.prepend_space_row(r1)
		program.remove(i)
		Result := true
	    end
	end
    end
end
