class SML
inherit
    TAG_SINGLETON;
    SIMPLIFY_WHITESPACE;
    REMOVE_TAGS
creation make
feature
    program : TWO_WAY_LINKED_LIST[ROW]

    make is
    do
	!!program.make
	init_tag_table
    end

    read is
    do
--std_error.put_string("parsing...")
	parse
--std_error.put_string("simplifying whitespace...")
	simplify_whitespace
--std_error.put_string("removing tags...")
	remove_tags
    end

    parse is
    local
	c : CHARACTER
    do
	!!current_row.make
	!!current_str.make_empty
	from
	    io.read_character
	    --speed optimization, can't reduce 0 bytes
	    if io.end_of_input then
		die_with_code(0)
	    end
	until io.end_of_input
	loop
	    c := io.last_character
	    if c = '<' then
		if not current_str.is_empty then
		    finish_row
		end
		parse_tag
	    elseif c = '>' then
		std_error.put_string("Unexpected '>'%N")
		die_with_code(1)
	    else
		add_char(c)
	    end
	    io.read_character
	end
	if not current_str.is_empty or
		not current_row.tag_list.is_empty then
	    finish_row
	end
    end

    sanity_check is
    local
	it : ITERATOR[ROW]
	it2 : ITERATOR[TAG_HOLDER]
	tag_stack : LINKED_LIST[TAG_HOLDER]
	tag : TAG
    do
	!!tag_stack.make
	it := program.get_new_iterator
	from it.start
	until it.is_off
	loop
	    it2 := it.item.tag_list.get_new_iterator
	    from it2.start
	    until it2.is_off
	    loop
		tag := it2.item.tag
		if tag.tag_start then
		    tag_stack.add_last(it2.item)
		else
		    if not (opposite_tag(tag) = tag_stack.last.tag) then
			std_error.put_string("bug: badly nested ")
			std_error.put_string(tag.name)
			std_error.put_new_line
			tag_stack.last.row.show_info
			it2.item.row.show_info
		    end
		    if it2.item.partner /= tag_stack.last
			or else tag_stack.last.partner /= it2.item then
			std_error.put_string("bug: bad partnerships: ")
			std_error.put_string(tag.name)
			std_error.put_new_line
			tag_stack.last.row.show_info
			it2.item.row.show_info
		    end
		    tag_stack.remove_last
		end
		if it2.item.row /= it.item then
		    std_error.put_string("bug: incorrect row info%N")
		    it.item.show_info
		    it2.item.row.show_info
		end
		if it2.item.redundant then
		    std_error.put_string("bug: redundancy remaining%N")
		    it.item.show_info
		    it2.item.row.show_info
		end
		it2.next
	    end
	    it.next
	end
    end

    current_row : ROW

    add_tag(t : TAG) is
    do
	current_row.add_tag(t)
    end

    --we keep a current_str here instead of going through current_row.text
    --every time we get a character
    current_str : STRING

    add_char(c : CHARACTER) is
    do
	current_str.add_last(c)
    end

    finish_row is
    do
	current_row.finish_row(current_str)
	program.add_last(current_row)
	!!current_row.make
	!!current_str.make(10)
	    --is make_empty better?
    end

    parse_tag is
    local
	word : STRING
    do
	!!word.make(2)
	from io.read_character
	until io.last_character = '>'
	loop
	    word.add_last(io.last_character)
	    io.read_character
	end
	if not tag_table.has(word) then
	    io.put_string("unimplemented tag: ")
	    io.put_string(word)
	    io.put_string(", sorry!%N")
	    die_with_code(1)
	else
	    add_tag(tag_table.at(word))
	end
    end

    output is
    local
	it : ITERATOR[ROW]
    do
	it := program.get_new_iterator
	from it.start
	until it.is_off
	loop
	    it.item.output
	    it.next
	end
    end

    show_info is
	--for debugging
    local
	it : ITERATOR[ROW]
    do
	it := program.get_new_iterator
	from it.start
	until it.is_off
	loop
	    it.item.show_info
	    it.next
	end
    end

    dump is
    local
	it : ITERATOR[ROW]
    do
	it := program.get_new_iterator
	from it.start
	until it.is_off
	loop
	    it.item.dump
	    it.next
	end
    end
end
